/* ********** Main ********** */
#connected-to-label {
  margin-top: 10px;
  margin-bottom: 10px;
}

#icon-button-box image {
  margin: 5px 2px;
}

#icon-button-box label {
  margin-left: 5px;
  margin-right: 3px;
}

/* ********** Playlist ********** */
#playlist-list-listbox row {
  margin: 0;
  padding: 0;
}

#playlist-list-spinner:checked,
#artist-list-spinner:checked {
  margin: 10px;
  padding: 0px;
}

#playlist-list-new-playlist-entry {
  margin: 10px 10px 5px 10px;
}

#playlist-list-new-playlist-cancel {
  margin: 5px 0 10px 0;
}

#playlist-list-new-playlist-confirm {
  margin: 5px 10px 10px 0;
}

#playlist-artwork-spinner, #artist-artwork-spinner, #album-artwork-spinner {
  min-height: 35px;
  min-width: 35px;
}

#playlist-album-artwork {
  min-height: 200px;
  min-width: 200px;
  margin: 10px;
}

#playlist-album-artwork.collapsed {
  min-height: 40px;
  min-width: 40px;
}

#playlist-name, #artist-detail-panel #artist-name {
  font-size: 40px;
  margin-bottom: 10px;
}

#playlist-comment, #playlist-stats, #artist-bio, #artist-stats, #similar-artists {
  margin-bottom: 10px;
}

#similar-artist-button {
  padding: 0;
  margin: -10px 0 0 10px;
}

#playlist-play-shuffle-buttons {
  margin-bottom: 10px;
}

/* ********** Playback Controls ********** */
#player-controls-album-artwork {
  min-height: 70px;
  min-width: 70px;
}

#player-controls-bar #play-button {
  min-height: 45px;
  min-width: 35px;
  border-width: 1px;
  border-radius: 45px;
}

/* Make the play icon look centered. */
#player-controls-bar #play-button image {
  margin-left: 5px;
  margin-right: 5px;
  margin-top: 1px;
  margin-bottom: 1px;
}

#player-controls-bar #song-scrubber {
  min-width: 400px;
}

#player-controls-bar #volume-slider {
  min-width: 90px;
}

#player-controls-bar #volume-slider value {
  margin-left: 8px;
}

#player-controls-bar #song-title {
  margin-bottom: 3px;
  font-weight: bold;
}

#player-controls-bar #album-name {
  margin-bottom: 3px;
  font-style: italic;
}

#up-next-popover #label {
  margin: 10px;
}

#play-queue-playing-icon {
  min-height: 40px;
  min-width: 40px;
}

#play-queue-row-image {
  min-height: 50px;
  min-width: 50px;
}

/* General */
.menu-button {
  padding: 5px;
}

/* ********** Artists & Albums ********** */
#grid-artwork-spinner, #album-list-song-list-spinner {
  min-height: 35px;
  min-width: 35px;
}

#grid-artwork {
  min-height: 200px;
  min-width: 200px;
  margin: 10px;
}

#grid-spinner {
  min-height: 50px;
  min-width: 50px;
  margin: 20px;
}

#grid-header-label {
  margin-left: 10px;
  margin-right: 10px;
  margin-bottom: 3px;
  font-weight: bold;
}

#grid-info-label {
  margin-left: 10px;
  margin-right: 10px;
  margin-bottom: 10px;
}

#artist-album-artwork {
  margin: 10px;
  min-width: 300px;
  min-height: 300px;
}

#artist-album-list-artwork {
  margin: 10px;
}

#artist-album-list-album-name {
  margin: 10px 10px 5px 10px;
  font-size: 25px;
}

#album-list-song-list-spinner {
  margin: 15px;
}
